@@ -1,7 +1,7 @@
 /*
  * nts_cookie.c - Network Time Security (NTS) cookie processing
- * Copyright 2019 by the NTPsec project contributors
- * SPDX-License-Identifier: BSD-4-Clause-UC
+ * Copyright the NTPsec project contributors
+ * SPDX-License-Identifier: BSD-2-Clause
  *
  * Section references are to
  * https://tools.ietf.org/html/draft-ietf-ntp-using-nts-for-ntp-15
@@ -28,7 +28,6 @@
 #include <pthread.h>
 #include <unistd.h>
 
-#include <openssl/rand.h>
 #include <aes_siv.h>
 
 #include "ntpd.h"
@@ -48,8 +47,19 @@
 
 /* K and I should be preserved across boots, and rotated every day or so.
  * We need to support the old K/I for another day.
- * Encryption within cookies uses AEAD_AES_SIV_CMAC_nnn.  That's the
- * same family of algorithims as NTS uses on the wire.
+ *
+ * If the file gets corrupted, blow it away and reboot.  It will get
+ * recreated, we will start using new cookies, packets from clients
+ * with old cookies will get dropped, and eventually clients will
+ * run out of cookies and use NTS-KE to get new ones.
+ *
+ * It would be possible to run without a cookie file.  Nobody would
+ * notice until the server was restarted.  Then there would be a flurry
+ * of NTS-KE requests until all clients obtained new/working cookies.
+ */
+
+/* Encryption within cookies uses AEAD_AES_SIV_CMAC_nnn.  That's the
+ * same family of algorithms as NTS uses on the wire.
  * The nnn is selected by the key length.
  *   32 => 256
  *   48 => 384
@@ -80,9 +90,9 @@
  * You can change that by editing the keys file.
  */
 int K_length = AEAD_AES_SIV_CMAC_256_KEYLEN;
-uint8_t K[NTS_MAX_KEYLEN], K2[NTS_MAX_KEYLEN];
-uint32_t I, I2;
 time_t K_time = 0;	/* time K was created, 0 for none */
+struct NTS_Key nts_keys[NTS_nKEYS];
+int nts_nKeys = 0;
 
 /* The mutex protects cookie_ctx
  * The NTS-KE servers can make cookies
@@ -94,7 +104,9 @@ AES_SIV_CTX* cookie_ctx;
 /* Statistics for ntpq */
 uint64_t nts_cookie_make = 0;
 uint64_t nts_cookie_decode = 0;
-uint64_t nts_cookie_decode_old = 0;
+uint64_t nts_cookie_decode_old = 0;	/* one day old */
+uint64_t nts_cookie_decode_old2 = 0;	/* two days old */
+uint64_t nts_cookie_decode_older = 0;	/* more than 2 days old */
 uint64_t nts_cookie_decode_too_old = 0;
 uint64_t nts_cookie_decode_error = 0;
 
@@ -120,15 +132,15 @@ bool nts_cookie_init(void) {
 bool nts_cookie_init2(void) {
 	bool OK = true;
 	if (!nts_read_cookie_keys()) {
-		nts_make_cookie_key();  /* make new cookie key */
-		nts_make_cookie_key();  /* push new to old, make new */
+		/* Can't read cookie file.  Make one */
+		nts_make_cookie_key();
 		K_time = time(NULL);
 		nts_write_cookie_keys();
 	}
 	return OK;
 }
 
-/* Rotate key -- 24 hours after last rotate
+/* Rotate key -- 24 hours after last rotation
  * That allows a cluster NTS-KE server to keep in sync
  * if we use ratchet rather than random.
  */
@@ -136,8 +148,9 @@ bool nts_cookie_init2(void) {
 // Set this shorter for debugging
 //  keys will timeout, packets will get dropped
 //  after 8 lost packets, it should go through the NTS-KE dance again
+// Just uncommenting the next line will generate a warning reminder.
 // #define SecondsPerDay 3600
-void nts_timer(void) {
+void nts_cookie_timer(void) {
 	time_t now;
 	if (0 == K_time) {
 		return;
@@ -152,9 +165,9 @@ void nts_timer(void) {
 		K_time += SecondsPerDay;
 	}
 	if (nts_write_cookie_keys() )
-		msyslog(LOG_INFO, "NTS: Wrote new cookie key.");
+		msyslog(LOG_INFO, "NTS: Wrote new cookie file, %d keys.", nts_nKeys);
 	else
-		msyslog(LOG_INFO, "NTS: Trouble writing new cookie key.");
+		msyslog(LOG_INFO, "NTS: Trouble writing new cookie file.");
 	return;
 }
 
@@ -185,39 +198,30 @@ bool nts_read_cookie_keys(void) {
 	if ( !((32 == K_length) || (48 == K_length) || (64 == K_length))) {
 		goto bail;
 	}
-	if (1 != fscanf(in, "I: %u\n", &I)) {
+	nts_nKeys = 0;
+	for (int i=0; i<NTS_nKEYS; i++) {
+	  struct NTS_Key *key = &nts_keys[i];
+	  if (1 != fscanf(in, "I: %u\n", &key->I)) {
+		if (0 < nts_nKeys) break;
 		goto bail;
-	}
-	if (0 != fscanf(in, "K: ")) {
+	  }
+	  if (0 != fscanf(in, "K: ")) {
 		goto bail;
-	}
-	for (int i=0; i< K_length; i++) {
+	  }
+	  for (int j=0; j< K_length; j++) {
 		unsigned int temp;
 		if (1 != fscanf(in, "%02x", &temp)) {
 			goto bail;
 		}
-		K[i] = temp;
-	}
-	if (0 != fscanf(in, "\n")) {
-		goto bail;
-	}
-	if (1 != fscanf(in, "I: %u\n", &I2)) {
-		goto bail;
-	}
-	if (0 != fscanf(in, "K: ")) {
-		goto bail;
-	}
-	for (int i=0; i< K_length; i++) {
-		unsigned int temp;
-		if (1 != fscanf(in, "%02x", &temp)) {
-			goto bail;
-		}
-		K2[i] = temp;
-	}
-	if (0 != fscanf(in, "\n")) {
+		key->K[j] = temp;
+	  }
+	  if (0 != fscanf(in, "\n")) {
 		goto bail;
+	  }
+	  nts_nKeys = i+1;
 	}
 	fclose(in);
+	msyslog(LOG_INFO, "NTS: Read cookie file, %d keys.", nts_nKeys);
 	return true;
 
   bail:
@@ -226,55 +230,65 @@ bool nts_read_cookie_keys(void) {
 	return false;
 }
 
-/* The draft describes a ratchet mode to make new keys
+/* RFC 8915 describes a ratchet mode to make new keys
  * That's one way to implement a KE server for a cluster of NTP servers.
  * The KE server and the NTP servers stay in sync without communication
  * after a one-time copy of the cookie file from NTP server to KE server.
+ * An alternative is to elect one system to generate new keys and
+ * they copy the key file to other systems and have them load it.
  */
-bool nts_make_cookie_key(void) {
-	bool OK = true;
-	memcpy(&K2, &K, sizeof(K2));	/* Push current cookie to old */
-	I2 = I;
-#if (OPENSSL_VERSION_NUMBER > 0x1010100fL)
-	OK &= RAND_priv_bytes(K, sizeof(K));
-#else
-	OK &= RAND_bytes(K, sizeof(K));
-#endif
-	OK &= RAND_bytes((uint8_t *)&I, sizeof(I));
-	return OK;
+void nts_make_cookie_key(void) {
+	if (nts_nKeys < NTS_nKEYS) nts_nKeys++;
+	for (int i=nts_nKeys-1; i>0; i--) {
+	  nts_keys[i] = nts_keys[i-1];
+	}
+	ntp_RAND_priv_bytes(nts_keys[0].K, K_length);
+	ntp_RAND_bytes((uint8_t *)&nts_keys[0].I, sizeof(nts_keys[0].I));
+	return;
 }
 
 bool nts_write_cookie_keys(void) {
-	const char *cookie_filename = NTS_COOKIE_KEY_FILE;
+	const char *cookiefile = NTS_COOKIE_KEY_FILE;
+	char tempfile[PATH_MAX];
 	int fd;
 	FILE *out;
 	char errbuf[100];
 	if (NULL != ntsconfig.KI)
-		cookie_filename = ntsconfig.KI;
-	fd = open(cookie_filename, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);
+		cookiefile = ntsconfig.KI;
+	strlcpy(tempfile, cookiefile, sizeof(tempfile));
+	strlcat(tempfile, "-tmp", sizeof(tempfile));
+	fd = open(tempfile, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);
 	if (-1 == fd) {
 		ntp_strerror_r(errno, errbuf, sizeof(errbuf));
-		msyslog(LOG_ERR, "ERR: can't open %s: %s", cookie_filename, errbuf);
+		msyslog(LOG_ERR, "ERR: can't open %s: %s", tempfile, errbuf);
 		return false;
 	}
 	out = fdopen(fd, "w");
 	if (NULL == out) {
 		ntp_strerror_r(errno, errbuf, sizeof(errbuf));
-		msyslog(LOG_ERR, "ERR: can't fdopen %s: %s", cookie_filename, errbuf);
+		msyslog(LOG_ERR, "ERR: can't fdopen %s: %s", tempfile, errbuf);
 		close(fd);
 		return false;
 	}
+
 	fprintf(out, "T: %lu\n", (unsigned long)K_time);
 	fprintf(out, "L: %d\n", K_length);
-	fprintf(out, "I: %u\n", I);
-	fprintf(out, "K: ");
-	for (int i=0; i< K_length; i++) fprintf(out, "%02x", K[i]);
-	fprintf(out, "\n");
-	fprintf(out, "I: %u\n", I2);
-	fprintf(out, "K: ");
-	for (int i=0; i< K_length; i++) fprintf(out, "%02x", K2[i]);
-	fprintf(out, "\n");
+	for (int i=0; i<nts_nKeys; i++) {
+	  struct NTS_Key *key = &nts_keys[i];
+	  fprintf(out, "I: %u\n", key->I);
+	  fprintf(out, "K: ");
+	    for (int j=0; j< K_length; j++) fprintf(out, "%02x", key->K[j]);
+	  fprintf(out, "\n");
+	  key++;
+	}
 	fclose(out);
+        if (rename(tempfile, cookiefile)) {
+	    ntp_strerror_r(errno, errbuf, sizeof(errbuf));
+            msyslog(LOG_WARNING,
+                    "LOG: Unable to rename temp cookie file %s to %s, %s",
+                    tempfile, cookiefile, errbuf);
+	    return false;
+	}
 	return true;
 }
 
@@ -314,11 +328,11 @@ int nts_make_cookie(uint8_t *cookie,
 	/* collect associated data */
 	finger = cookie;
 
-	memcpy(finger, &I, sizeof(I));
-	finger += sizeof(I);
+	memcpy(finger, &nts_keys[0].I, sizeof(nts_keys[0].I));
+	finger += sizeof(nts_keys[0].I);
 
 	nonce = finger;
-	RAND_bytes(finger, NONCE_LENGTH);
+	ntp_RAND_bytes(finger, NONCE_LENGTH);
 	finger += NONCE_LENGTH;
 
 	used = finger-cookie;
@@ -328,7 +342,7 @@ int nts_make_cookie(uint8_t *cookie,
 
 	ok = AES_SIV_Encrypt(cookie_ctx,
 			     finger, &left,   /* left: in: max out length, out: length used */
-			     K, K_length,
+			     nts_keys[0].K, K_length,
 			     nonce, NONCE_LENGTH,
 			     plaintext, plainlength,
 			     cookie, AD_LENGTH);
@@ -357,32 +371,54 @@ bool nts_unpack_cookie(uint8_t *cookie, int cookielen,
   uint8_t *c2s, uint8_t *s2c, int *keylen) {
 	uint8_t *finger;
 	uint8_t plaintext[NTS_MAX_COOKIELEN];
-	uint8_t *key;
 	uint8_t *nonce;
 	uint32_t temp;
 	size_t plainlength;
 	int cipherlength;
 	bool ok;
+	struct NTS_Key *key;
+	int i;
 
 	if (NULL == cookie_ctx)
 		return false;	/* We aren't initialized yet. */
 
+	if (0 == nts_nKeys) {
+		return false;  /* We are not a NTS enabled server. */
+	}
+
 	/* We may get garbage from the net */
 	if (cookielen > NTS_MAX_COOKIELEN)
 		return false;
 
 	finger = cookie;
-	if (0 == memcmp(finger, &I, sizeof(I))) {
-		key = K;
+	key = NULL;		/* squash uninitialized warning */
+	for (i=0; i<nts_nKeys; i++) {
+	  key = &nts_keys[i];
+	  if (0 == memcmp(finger, &key->I, sizeof(key->I))) {
+		break;
+	  }
+	}
+	if (0 == i) {
 		nts_cookie_decode++;
-	} else if (0 == memcmp(finger, &I2, sizeof(I2))) {
-		key = K2;
-		nts_cookie_decode_old++;
-	} else {
+	} else if (nts_nKeys == i) {
 		nts_cookie_decode_too_old++;
 		return false;
+	} else if (1 == i) {
+		nts_cookie_decode_old++;
+	} else if (2 == i) {
+		nts_cookie_decode_old2++;
+	} else {
+		nts_cookie_decode_older++;
 	}
-	finger += sizeof(I);
+#if 0
+	if (1<i) {
+	  /* Hack for debugging */
+	  /* Beware: DoS possibility on a public server */
+	  msyslog(LOG_INFO, "NTS: Old cookie: %d days.", i);
+	}
+#endif
+
+	finger += sizeof(key->I);
 	nonce = finger;
 	finger += NONCE_LENGTH;
 
@@ -395,7 +431,7 @@ bool nts_unpack_cookie(uint8_t *cookie, int cookielen,
 
 	ok = AES_SIV_Decrypt(cookie_ctx,
 			     plaintext, &plainlength,
-			     key, K_length,
+			     key->K, K_length,
 			     nonce, NONCE_LENGTH,
 			     finger, cipherlength,
 			     cookie, AD_LENGTH);
